class Logger:
    """Logger is the logging interface implementation."""

    def enable_log(self, enable):
        """controls whether print the message."""
        pass

    def is_enabled(self):
        """returns if logger is enabled."""
        pass

    def write(self, *v):
        """formats using the default formats for its operands and logs the message."""
        pass

    def writef(self, fmt, *v):
        """formats according to a format specifier and logs the message."""
        pass



class Effector:
    """Effector is the interface for Casbin effectors."""

    ALLOW = 0

    INDETERMINATE = 1

    DENY = 2

    def merge_effects(self, expr, effects, results):
        """merges all matching results collected by the enforcer into a single decision."""
        pass



import fnmatch
import re
import ipaddress

KEY_MATCH2_PATTERN = re.compile(r'(.*?):[^\/]+(.*?)')
KEY_MATCH3_PATTERN = re.compile(r'(.*?){[^\/]+}(.*?)')


def key_match(key1, key2):
    """determines whether key1 matches the pattern of key2 (similar to RESTful path), key2 can contain a *.
    For example, "/foo/bar" matches "/foo/*"
    """

    i = key2.find("*")
    if i == -1:
        return key1 == key2

    if len(key1) > i:
        return key1[:i] == key2[:i]
    return key1 == key2[:i]


def key_match_func(*args):
    """The wrapper for key_match.
    """
    name1 = args[0]
    name2 = args[1]

    return key_match(name1, name2)


def key_match2(key1, key2):
    """determines whether key1 matches the pattern of key2 (similar to RESTful path), key2 can contain a *.
    For example, "/foo/bar" matches "/foo/*", "/resource1" matches "/:resource"
    """

    key2 = key2.replace("/*", "/.*")
    key2 = KEY_MATCH2_PATTERN.sub(r'\g<1>[^\/]+\g<2>', key2, 0)

    return regex_match(key1, "^" + key2 + "$")


def key_match2_func(*args):
    name1 = args[0]
    name2 = args[1]

    return key_match2(name1, name2)


def key_match3(key1, key2):
    """determines determines whether key1 matches the pattern of key2 (similar to RESTful path), key2 can contain a *.
    For example, "/foo/bar" matches "/foo/*", "/resource1" matches "/{resource}"
    """

    key2 = key2.replace("/*", "/.*")
    key2 = KEY_MATCH3_PATTERN.sub(r'\g<1>[^\/]+\g<2>', key2, 0)

    return regex_match(key1, "^" + key2 + "$")


def key_match3_func(*args):
    name1 = args[0]
    name2 = args[1]

    return key_match3(name1, name2)


def regex_match(key1, key2):
    """determines whether key1 matches the pattern of key2 in regular expression."""

    res = re.match(key2, key1)
    if res:
        return True
    else:
        return False


def regex_match_func(*args):
    """the wrapper for RegexMatch."""

    name1 = args[0]
    name2 = args[1]

    return regex_match(name1, name2)


def glob_match(string, pattern):
    """determines whether string matches the pattern in glob expression."""
    return fnmatch.fnmatch(string, pattern)


def glob_match_func(*args):
    """the wrapper for globMatch."""

    string = args[0]
    pattern = args[1]

    return glob_match(string, pattern)


def ip_match(ip1, ip2):
    """IPMatch determines whether IP address ip1 matches the pattern of IP address ip2, ip2 can be an IP address or a CIDR pattern.
    For example, "192.168.2.123" matches "192.168.2.0/24"
    """
    ip1 = ipaddress.ip_address(ip1)
    try:
        network = ipaddress.ip_network(ip2, strict=False)
        return ip1 in network
    except ValueError:
        return ip1 == ip2


def ip_match_func(*args):
    """the wrapper for IPMatch."""

    ip1 = args[0]
    ip2 = args[1]

    return ip_match(ip1, ip2)


def generate_g_function(rm):
    """the factory method of the g(_, _) function."""

    def f(*args):
        name1 = args[0]
        name2 = args[1]

        if not rm:
            return name1 == name2
        elif 2 == len(args):
            return rm.has_link(name1, name2)
        else:
            domain = str(args[2])
            return rm.has_link(name1, name2, domain)

    return f



from .default_logger import DefaultLogger

logger = DefaultLogger()


def set_logger(l):
    """sets the current logger."""
    global logger
    logger = l


def get_logger():
    """returns the current logger."""
    return logger


def log_print(*v):
    """prints the log."""
    logger.write(*v)


def log_printf(fmt, *v):
    """prints the log with the format."""
    logger.writef(fmt, *v)



from .logger import Logger
import logging


class DefaultLogger(Logger):
    """the implementation for a Logger using logging."""

    enable = False
    def __init__(self):
        self.logger = logging.getLogger('casbin')
        self.logger.setLevel(logging.INFO)
        handler = logging.StreamHandler()
        fmt = "%(asctime)s - %(levelname)s - %(message)s"
        formatter = logging.Formatter(fmt)
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)

    def enable_log(self, enable):
        """controls whether print the message."""
        self.enable = enable

    def is_enabled(self):
        """returns if logger is enabled."""
        return self.enable

    def write(self, *v):
        """formats using the default formats for its operands and logs the message."""
        if self.enable:
            s = ""
            for vv in v:
                s = s + str(vv)
            self.logger.info(s)

    def writef(self, fmt, *v):
        """formats according to a format specifier and logs the message."""
        if self.enable:
            self.logger.info(fmt, *v)



def load_policy_line(line, model):
    """loads a text line as a policy rule to model."""

    if line == "":
        return

    if line[:1] == "#":
        return

    tokens = line.split(", ")
    key = tokens[0]
    sec = key[0]

    if sec not in model.model.keys():
        return

    if key not in model.model[sec].keys():
        return

    model.model[sec][key].policy.append(tokens[1:])


class Adapter:
    """the interface for Casbin adapters."""

    def load_policy(self, model):
        """loads all policy rules from the storage."""
        pass

    def save_policy(self, model):
        """saves all policy rules to the storage."""
        pass

    def add_policy(self, sec, ptype, rule):
        """adds a policy rule to the storage."""
        pass

    def remove_policy(self, sec, ptype, rule):
        """removes a policy rule from the storage."""
        pass

    def remove_filtered_policy(self, sec, ptype, field_index, *field_values):
        """removes policy rules that match the filter from the storage.
        This is part of the Auto-Save feature.
        """
        pass



from casbin import util, log


class Policy:
    def __init__(self):
        self.model = {}

    def build_role_links(self, rm):
        """initializes the roles in RBAC."""

        if "g" not in self.model.keys():
            return

        for ast in self.model["g"].values():
            ast.build_role_links(rm)

    def print_policy(self):
        """prints the policy to log."""

        log.log_print("Policy:")
        for sec in ["p", "g"]:
            if sec not in self.model.keys():
                continue

            for key, ast in self.model[sec].items():
                log.log_print(key, ": ", ast.value, ": ", ast.policy)

    def clear_policy(self):
        """clears all current policy."""

        for sec in ["p", "g"]:
            if sec not in self.model.keys():
                continue

            for key, ast in self.model[sec].items():
                self.model[sec][key].policy = []

    def get_policy(self, sec, ptype):
        """gets all rules in a policy."""

        return self.model[sec][ptype].policy

    def get_filtered_policy(self, sec, ptype, field_index, *field_values):
        """gets rules based on field filters from a policy."""
        res = []

        for rule in self.model[sec][ptype].policy:
            matched = True
            for i, field_value in enumerate(field_values):
                if field_value != '' and rule[field_index + i] != field_value:
                    matched = False
                    break

            if matched:
                res.append(rule)

        return res

    def has_policy(self, sec, ptype, rule):
        """determines whether a model has the specified policy rule."""
        if sec not in self.model.keys():
            return False
        if ptype not in self.model[sec]:
            return False

        return rule in self.model[sec][ptype].policy

    def add_policy(self, sec, ptype, rule):
        """adds a policy rule to the model."""

        if not self.has_policy(sec, ptype, rule):
            self.model[sec][ptype].policy.append(rule)
            return True

        return False

    def add_policies(self,sec,ptype,rules):
        """adds policy rules to the model."""
        
        for rule in rules:
            if self.has_policy(sec,ptype,rule):
                return False

        for rule in rules:
            self.model[sec][ptype].policy.append(rule)
        
        return True

    def remove_policy(self, sec, ptype, rule):
        """removes a policy rule from the model."""

        if sec not in self.model.keys():
            return False
        if ptype not in self.model[sec]:
            return False

        if not self.has_policy(sec, ptype, rule):
            return False

        self.model[sec][ptype].policy.remove(rule)

        return rule not in self.model[sec][ptype].policy

    def remove_policies(self, sec, ptype, rules):
        """RemovePolicies removes policy rules from the model."""

        if sec not in self.model.keys():
            return False
        if ptype not in self.model[sec]:
            return False

        for rule in rules:
            if not self.has_policy(sec,ptype,rule):
                return False
            self.model[sec][ptype].policy.remove(rule)
            if rule in self.model[sec][ptype].policy:
                return False
        
        return True

    def remove_filtered_policy(self, sec, ptype, field_index, *field_values):
        """removes policy rules based on field filters from the model."""
        tmp = []
        res = False

        if sec not in self.model.keys():
            return res
        if ptype not in self.model[sec]:
            return res

        for rule in self.model[sec][ptype].policy:
            matched = True
            for i, field_value in enumerate(field_values):
                if field_value != '' and rule[field_index + i] != field_value:
                    matched = False
                    break

            if matched:
                res = True
            else:
                tmp.append(rule)

        self.model[sec][ptype].policy = tmp

        return res

    def get_values_for_field_in_policy(self, sec, ptype, field_index):
        """gets all values for a field for all rules in a policy, duplicated values are removed."""

        values = []
        if sec not in self.model.keys():
            return values
        if ptype not in self.model[sec]:
            return values

        for rule in self.model[sec][ptype].policy:
            values.append(rule[field_index])

        return util.array_remove_duplicates(values)



from casbin import log


class Assertion:
    def __init__(self):
        self.key = ""
        self.value = ""
        self.tokens = []
        self.policy = []
        self.rm = None

    def build_role_links(self, rm):
        self.rm = rm
        count = self.value.count("_")

        for rule in self.policy:
            if count < 2:
                raise RuntimeError('the number of "_" in role definition should be at least 2')

            if len(rule) < count:
                raise RuntimeError("grouping policy elements do not meet role definition")

            self.rm.add_link(*rule[:count])

        log.log_print("Role links for: " + self.key)
        self.rm.print_roles()



class RoleManager:
    """provides interface to define the operations for managing roles."""

    def clear(self):
        pass

    def add_link(self, name1, name2, *domain):
        pass

    def delete_link(self, name1, name2, *domain):
        pass

    def has_link(self, name1, name2, *domain):
        pass

    def get_roles(self, name, *domain):
        pass

    def get_users(self, name, *domain):
        pass

    def print_roles(self):
        pass



from io import StringIO


class Config:
    """represents an implementation of the ConfigInterface"""

    # DEFAULT_SECTION specifies the name of a section if no name provided
    DEFAULT_SECTION = 'default'
    # DEFAULT_COMMENT defines what character(s) indicate a comment `#`
    DEFAULT_COMMENT = '#'
    # DEFAULT_COMMENT_SEM defines what alternate character(s) indicate a comment `;`
    DEFAULT_COMMENT_SEM = ';'
    # DEFAULT_MULTI_LINE_SEPARATOR defines what character indicates a multi-line content
    DEFAULT_MULTI_LINE_SEPARATOR = '\\'

    _data = dict()

    def __init__(self):
        self._data = dict()

    @staticmethod
    def new_config(conf_name):
        c = Config()
        c._parse(conf_name)
        return c

    @staticmethod
    def new_config_from_text(text):
        c = Config()
        f = StringIO(text)
        c._parse_buffer(f)
        return c

    def add_config(self, section, option, value):
        if section == '':
            section = self.DEFAULT_SECTION

        if section not in self._data.keys():
            self._data[section] = {}

        self._data[section][option] = value

    def _parse(self, fname):
        with open(fname, 'r', encoding='utf-8') as f:
            self._parse_buffer(f)

    def _parse_buffer(self, f):
        section = ''
        line_num = 0
        buf = []
        can_write = False
        while True:
            if can_write:
                self._write(section, line_num, buf)
                can_write = False
            line_num = line_num + 1

            line = f.readline()

            if not line:
                if len(buf) > 0:
                    self._write(section, line_num, buf)
                break
            line = line.strip()

            if '' == line or self.DEFAULT_COMMENT == line[0:1] or self.DEFAULT_COMMENT_SEM == line[0:1]:
                can_write = True
                continue
            elif '[' == line[0:1] and ']' == line[-1]:
                if len(buf) > 0:
                    self._write(section, line_num, buf)
                    can_write = False
                section = line[1:-1]
            else:
                p = ''
                if self.DEFAULT_MULTI_LINE_SEPARATOR == line[-1]:
                    p = line[0:-1].strip()
                    p = p + ' ' 
                else:
                    p = line
                    can_write = True
                buf.append(p)

    def _write(self, section, line_num, b):

        buf = "".join(b)
        if len(buf) <= 0:
            return
        option_val = buf.split('=', 1)

        if len(option_val) != 2:
            raise RuntimeError('parse the content error : line {} , {} = ?'.format(line_num, option_val[0]))

        option = option_val[0].strip()
        value = option_val[1].strip()

        self.add_config(section, option, value)

        del b[:]

    def set(self, key, value):
        if len(key) == 0:
            raise RuntimeError("key is empty")

        keys = key.lower().split('::')
        if len(keys) >= 2:
            section = keys[0]
            option = keys[1]
        else:
            section = ""
            option = keys[0]
        self.add_config(section, option, value)

    def get(self, key):
        """section.key or key"""

        keys = key.lower().split('::')
        if len(keys) >= 2:
            section = keys[0]
            option = keys[1]
        else:
            section = self.DEFAULT_SECTION
            option = keys[0]

        if section in self._data.keys():
            if option in self._data[section].keys():
                return self._data[section][option]
        return ''



from .effector import Effector


class DefaultEffector(Effector):
    """default effector for Casbin."""

    def merge_effects(self, expr, effects, results):
        """merges all matching results collected by the enforcer into a single decision."""

        effects = set(effects)
        result = False
        if expr == "some(where (p_eft == allow))":
            if self.ALLOW in effects:
                result = True

        elif expr == "!some(where (p_eft == deny))":
            result = True

            if self.DENY in effects:
                result = False

        elif expr == "some(where (p_eft == allow)) && !some(where (p_eft == deny))":
            if self.DENY in effects:
                result = False
            elif self.ALLOW in effects:
                result = True

        elif expr == "priority(p_eft) || deny":
            for eft in effects:
                if eft != self.INDETERMINATE:
                    if eft == self.ALLOW:
                        result = True
                    else:
                        result = False
                    break
        else:
            raise RuntimeError("unsupported effect")

        return result



from unittest import TestCase
from casbin import log


class TestLog(TestCase):

    def test_log(self):
        log.get_logger().enable_log(True)
        log.log_print("test log", "print")
        log.log_printf("test log %s", "print")

        self.assertTrue(log.get_logger().is_enabled())



import setuptools

desc_file = "README.md"

with open(desc_file, "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="casbin",
    author="TechLee",
    author_email="techlee@qq.com",
    description="An authorization library that supports access control models like ACL, RBAC, ABAC in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/casbin/pycasbin",
    keywords=["casbin", "acl", "rbac", "abac", "auth", "authz", "authorization", "access control", "permission"],
    packages=setuptools.find_packages(exclude=("tests",)),
    install_requires=['simpleeval>=0.9.10'],
    python_requires=">=3.3",
    license="Apache 2.0",
    classifiers=[
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    data_files=[desc_file],
)


from casbin import persist
from casbin import model
from .file_adapter import FileAdapter
import os

class Filter:
    #P,G are string []
    P = []
    G = []

class FilteredAdapter (FileAdapter,persist.FilteredAdapter):
    filtered = False
    _file_path = ""
    filter = Filter()
    #new_filtered_adapte is the constructor for FilteredAdapter.
    def __init__(self,file_path):
        self.filtered = True
        self._file_path = file_path
    
    def load_policy(self,model):
        if not os.path.isfile(self._file_path):
            raise RuntimeError("invalid file path, file path cannot be empty")
        self.filtered=False
        self._load_policy_file(model)

    #load_filtered_policy loads only policy rules that match the filter.
    def load_filtered_policy(self,model,filter):
        if filter == None:
            return self.load_policy(model)
        
        if not os.path.isfile(self._file_path):
            raise RuntimeError("invalid file path, file path cannot be empty")

        try:
            filter_value = [filter.__dict__['P']]+[filter.__dict__['G']]
        except:
            raise RuntimeError("invalid filter type")

        self.load_filtered_policy_file(model,filter_value,persist.load_policy_line)
        self.filtered = True

    def load_filtered_policy_file(self,model,filter,hanlder):
        with open(self._file_path, "rb") as file:
            while True:
                line = file.readline()
                line = line.decode().strip()
                if line == '\n':
                    continue
                if not line :
                    break
                if filter_line(line,filter):
                    continue
        
                hanlder(line,model)

    #is_filtered returns true if the loaded policy has been filtered.
    def is_filtered(self):
        return self.filtered
    def save_policy(self,model):
        if self.filtered:
            raise RuntimeError("cannot save a filtered policy")

        self._save_policy_file(model)

def filter_line(line,filter):
    if filter == None:
        return False
    
    p = line.split(',')
    if len(p) == 0:
        return True
    filter_slice = []

    if  p[0].strip()== 'p':
        filter_slice = filter[0]
    elif p[0].strip() == 'g':
        filter_slice = filter[1]
    return filter_words(p,filter_slice)

def filter_words(line,filter):
    if len(line) < len(filter)+1:
        return True
    skip_line=False
    for i,v in enumerate(filter):
        if(len(v) >0 and ( v.strip() != line[i+1].strip() ) ):
            skip_line = True
            break

    return skip_line 